pub mod job_config;
mod swarm;

pub use swarm::swarm_config;

use job_config::{Bounds, JobConfig, StopCondition};
use std::thread;
use swarm::{GlobalRecord, Swarm};
use swarm_config::SwarmConfig;

/// Particle Swarm Optimizer
///
/// Base struct to run optimization jobs using Particle Swarms and an objective function
///
#[derive(Clone, Debug)]
pub struct PSO {
    num_swarms: usize,
    swarm_configs: Vec<SwarmConfig>,
    verbose: bool,
}

impl PSO {
    /// Generate a PSO from a list of Swarm Configuartions
    pub fn from_swarm_configs(verbose: bool, swarm_configs: Vec<SwarmConfig>) -> Self {
        let num_swarms = swarm_configs.len();

        assert_ne!(num_swarms, 0, "At least One SwarmConfig must be provided!");

        if verbose {
            initialization_message(num_swarms);

            println!("\nSwarm Configurations: ");
            for (swarm_index, sc) in swarm_configs.iter().enumerate() {
                println!("Swarm");
                cyan_ln!("{}:", swarm_index);
                green_ln!("{}", sc);
            }
        }

        Self {
            num_swarms,
            swarm_configs,
            verbose,
        }
    }

    /// Generate a PSO with `num_swarms` swarms using the given Swarm Configuration
    pub fn from_swarm_config(num_swarms: usize, verbose: bool, swarm_config: &SwarmConfig) -> Self {
        assert_ne!(0, num_swarms, "The PSO must have at least one Swarm!");

        let swarm_configs = (0..num_swarms).map(|_| swarm_config.clone()).collect();
        if verbose {
            initialization_message(num_swarms);

            println!("\nSwarm Configuration: ");
            green_ln!("{}", swarm_config);
        }

        Self {
            num_swarms,
            swarm_configs,
            verbose,
        }
    }

    /// Generate a PSO with `num_swarms` swarms using the default Swarm Configuration
    pub fn default(num_swarms: usize, verbose: bool) -> Self {
        let swarm_config = SwarmConfig::new();

        Self::from_swarm_config(num_swarms, verbose, &swarm_config)
    }

    /// Run an optimization with the given Job Configuration and Objective Function
    ///
    /// Returns a tuple with the minimum cost and corresponding location in the search space
    ///
    /// This function clones the objective function for each swarm before sharing it across threads.
    /// This is a performance consideration to reduce the shared use of data within the objective function across multiple threads.
    /// It also allows the objective function to internally mutate itself (the function does not have to be Sync+Send)
    ///
    pub fn run_job<F>(&self, job_config: JobConfig, objective_function: F) -> (f64, Vec<f64>)
    where
        F: Fn(&[f64]) -> f64 + Clone + Sync + Send + 'static,
    {
        println!("\n Starting Optimization Job with: ");
        dark_cyan_ln!("{}", job_config);

        let global_record = GlobalRecord::new(job_config.num_variables);

        let swarm_threads: Vec<std::thread::JoinHandle<()>> = self
            .swarm_configs
            .iter()
            .enumerate()
            .map(|(i, swarm_config)| {
                let ob = objective_function.clone();
                let jc = job_config.clone();
                let sc = swarm_config.clone();
                let mut gr = global_record.clone();

                thread::Builder::new()
                    .name(format!("swarm_thread_{}", i))
                    .spawn(move || {
                        let mut swarm = Swarm::new(format!("swarm_{}", i), sc, &jc);
                        swarm.minimize(ob, &mut gr)
                    })
                    .unwrap()
            })
            .collect();

        for st in swarm_threads {
            st.join().expect("Thread panicked during operation!");
        }

        let min = global_record
            .to_tuple()
            .expect("Should not block, main thread is the only owner here");

        print!("minimum of: ");
        green!("{} ", min.0);
        print!("located at: ");
        blue_ln!("{:?}", min.1);

        min
    }
}

fn initialization_message(num_swarms: usize) {
    if num_swarms == 1 {
        print!("PSO Initialized with ");
        yellow!("1");
        println!(" Swarm!");
    } else {
        print!("PSO Initialized with ");
        yellow!("{}", num_swarms);
        println!(" Swarms!");
    }
}
