use rand::distributions::{Distribution, Uniform};
use rand::prelude::ThreadRng;
use std::fmt;

const DEFAULT_MAX_VEL: f64 = 2.5;
const DEFAULT_VAR_BOUND: f64 = 1.0;
const DEFAULT_NUM_ITERATIONS: usize = 1000;
const DEFAULT_CONSOLE_UPDATE: usize = 1000;

/// Configuration for one Optimization Job
///
/// * Defines the number of variables, or search space dimensions
/// * Defines the stop condition for the optimization
/// * Defines the boundaries on the search space
/// * Defines the velocity limits along each variable of the search space
///
/// # Examples
///
/// ```
/// use pso::job_config::JobConfig;
///
/// // Start with the defualt values and a 3 dimensional search space
/// let mut jc = JobConfig::new(3);
///
/// // Change settings as desired
/// jc.exit_cost(5e-10);
/// jc.variable_bounds(vec![[-1.0, 1.0], [0.0, 5.0], [-100.0, 100.0]]);
/// jc.max_velocity(1.25);
///
/// println!("{}", jc);
/// ```
///
#[derive(Clone, Debug)]
pub struct JobConfig {
    pub variable_bounds: VariableBounds,
    pub velocity_bounds: VelocityBounds,
    pub stop_condition: StopCondition,
    pub update_console: Option<usize>,
    pub num_variables: usize,
}

impl JobConfig {
    /// Constructs a default job configuration for a `num_variables` dimension search space
    pub fn new(num_variables: usize) -> Self {
        Self {
            variable_bounds: VariableBounds::default(num_variables),
            velocity_bounds: VelocityBounds::default(num_variables),
            stop_condition: StopCondition::Iterations(DEFAULT_NUM_ITERATIONS),
            update_console: Some(DEFAULT_CONSOLE_UPDATE),
            num_variables,
        }
    }

    /// Sets the stop condition to end the job after `max_iterations` iterations
    pub fn max_iterations(&mut self, max_iterations: usize) {
        self.stop_condition = StopCondition::Iterations(max_iterations);
    }

    /// Sets the stop condition to end the job once the minimum cost drops below `minimum_cost`
    pub fn exit_cost(&mut self, minimum_cost: f64) {
        self.stop_condition = StopCondition::Cost(minimum_cost);
    }

    /// Sets the stop condition to a union of [`max_iterations()`] and [`exit_cost()`]
    ///     The optimization job will run for `max_iterations` iterations
    ///     unless the minimum cost drops below `minimum_cost` first.
    pub fn max_iter_and_exit_cost(&mut self, max_iterations: usize, exit_cost: f64) {
        self.stop_condition = StopCondition::Both(max_iterations, exit_cost);
    }

    /// Sets the bounds on all variables to `variable_bound`
    ///     Where `variable_bound[0]` is the lower bound and `variable_bound[1]` is the upper bound
    ///
    /// # Panics
    /// * If the first element of any of the bounds is larger than the second element
    ///
    pub fn variable_bound(&mut self, variable_bound: [f64; 2]) {
        self.variable_bounds = VariableBounds::from_range(self.num_variables, variable_bound);
    }

    /// Sets the variable bounds individualy for each variable in the search space
    ///
    /// # Panics
    /// * If the length of `variable_bounds` is not equal to the number of dimensions in the search space
    /// * FI the first element of any of the bounds is larger than the second element
    ///
    pub fn variable_bounds(&mut self, variable_bounds: Vec<[f64; 2]>) {
        assert_eq!(
            self.num_variables,
            variable_bounds.len(),
            "Variable Bounds Vector must have one element for each variable!"
        );
        self.variable_bounds = VariableBounds::from_ranges(variable_bounds);
    }

    /// Sets the maximum velocity along each dimension to `max_velocity`
    ///
    /// # Panics
    /// * If `max_velocity` is a negative number
    ///
    pub fn max_velocity(&mut self, max_velocity: f64) {
        self.velocity_bounds = VelocityBounds::from_max(self.num_variables, max_velocity);
    }

    /// Sets the Maximum Velocites along each dimension individually
    ///
    /// # Panics
    /// * If any of the Max Velocities are negative numbers
    /// * If the length of `max_velocities` is not equal to the number of dimensions in the search space
    ///
    pub fn max_velocities(&mut self, max_velocities: Vec<f64>) {
        assert_eq!(
            self.num_variables,
            max_velocities.len(),
            "Max Velocites Vector must have one element for each variable!"
        );
        self.velocity_bounds = VelocityBounds::from_maxs(max_velocities);
    }

    /// Set the number of iterations between console updates for each swarm
    ///
    /// Supply None to prevent console updates
    pub fn update_console(&mut self, update_rate: Option<usize>) {
        assert_ne!(update_rate, Some(0), "Update rate must not be zero!");
        self.update_console = update_rate;
    }
}

impl fmt::Display for JobConfig {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(
            f,
            "Optimization Job over {} variables \n {} \n {} \n {}",
            self.num_variables, self.variable_bounds, self.velocity_bounds, self.stop_condition
        )
    }
}

#[derive(Clone, Debug)]
pub enum StopCondition {
    Iterations(usize),
    Cost(f64),
    Both(usize, f64),
}

impl fmt::Display for StopCondition {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Optimizer will stop running ")?;
        match self {
            Self::Iterations(num_iterations) => write!(f, "after {} iterations.", num_iterations),
            Self::Cost(stop_cost) => {
                write!(f, "when the minimum cost drops below {:e}.", stop_cost)
            }
            Self::Both(num_iterations, stop_cost) => write!(
                f,
                "after {} iterations, unless the minimum cost drops below {:e} first.",
                num_iterations, stop_cost
            ),
        }
    }
}

pub trait Bounds {
    fn default(num_variables: usize) -> Self;
    fn sample_bounds(&self, rng: &mut ThreadRng) -> Vec<f64>;
    fn bound_array(&self) -> Vec<[f64; 2]>;
}

#[derive(Clone, Debug)]
pub struct VariableBounds {
    ranges: Vec<[f64; 2]>,
    dists: Vec<Uniform<f64>>,
}

impl VariableBounds {
    pub fn from_ranges(ranges: Vec<[f64; 2]>) -> Self {
        let dists = ranges.iter().map(|r| Uniform::new(r[0], r[1])).collect();
        Self {
            ranges: ranges.clone(),
            dists,
        }
    }

    pub fn from_range(num_variables: usize, range: [f64; 2]) -> Self {
        let ranges = vec![range; num_variables];
        let dist = Uniform::new(range[0], range[1]);
        let dists = vec![dist; num_variables];

        Self { ranges, dists }
    }
}

impl Bounds for VariableBounds {
    fn default(num_variables: usize) -> Self {
        Self::from_range(num_variables, [-1.0 * DEFAULT_VAR_BOUND, DEFAULT_VAR_BOUND])
    }

    fn sample_bounds(&self, rng: &mut ThreadRng) -> Vec<f64> {
        self.dists.iter().map(|dist| dist.sample(rng)).collect()
    }

    fn bound_array(&self) -> Vec<[f64; 2]> {
        self.ranges.clone()
    }
}

impl fmt::Display for VariableBounds {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Variable Bounds: [ ")?;
        for range in &self.ranges {
            write!(f, "({} <--> {}) ", range[0], range[1])?;
        }
        write!(f, "]")
    }
}

#[derive(Clone, Debug)]
pub struct VelocityBounds {
    maxs: Vec<f64>,
    dists: Vec<Uniform<f64>>,
}

impl VelocityBounds {
    pub fn from_maxs(maxs: Vec<f64>) -> Self {
        let dists = maxs
            .iter()
            .map(|m| {
                assert!(m > &0.0, "All Max Velocities must be greater than zero!");
                Uniform::new(-1.0 * m, m)
            })
            .collect();

        Self {
            maxs: maxs.clone(),
            dists,
        }
    }

    pub fn from_max(num_variables: usize, max: f64) -> Self {
        assert!(max > 0.0, "Max Velocity must be greater than Zero!");
        let maxs = vec![max; num_variables];
        let dist = Uniform::new(-1.0 * max, max);
        let dists = vec![dist; num_variables];

        Self { maxs, dists }
    }
}

impl Bounds for VelocityBounds {
    fn default(num_variables: usize) -> Self {
        Self::from_max(num_variables, DEFAULT_MAX_VEL)
    }

    fn sample_bounds(&self, rng: &mut ThreadRng) -> Vec<f64> {
        self.dists.iter().map(|dist| dist.sample(rng)).collect()
    }

    fn bound_array(&self) -> Vec<[f64; 2]> {
        self.maxs.iter().map(|max| [-1.0 * max, *max]).collect()
    }
}

impl fmt::Display for VelocityBounds {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Max Velocities: [")?;
        for max in &self.maxs {
            write!(f, "{}, ", max)?;
        }
        write!(f, "]")
    }
}
